/*
** Module   : PushMessagesWorker.java
** Abstract : a thread used to push messages to the remote end point.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20140220 Created initial version.
*/

package com.goldencode.p2j.ui.client.chui.driver.web;

import java.nio.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;

import org.eclipse.jetty.websocket.api.*;

import com.goldencode.p2j.util.*;

/**
 * For performances reasons the messages send to the remote end point via web sockets
 * should be sent asynchronously. A queue is used to store messages. Messages are
 * extracted from queue on a FIFO order and pushed down to the remote end point.
 * 
 * @author mag
 *
 */
public class PushMessagesWorker 
extends Thread
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(PushMessagesWorker.class.getName()); 
   
   /** time to wait for thread exit */
   private static final long WAIT_TO_DIE = 5000;

   /** Keyboard input queue */
   private volatile Deque<Object> messages = new ConcurrentLinkedDeque<>();
   
   /** Remote session */
   private volatile Session session;  

   /** Remote endpoint */
   private volatile RemoteEndpoint remote;  
   
   /** Running flag */
   private volatile boolean running = true;
   
   /**
    * Constructor.
    * 
    * @param   session
    *          Remote end point connection.
    */
   public PushMessagesWorker(Session session)
   {
      this.setName("pushworker");
      this.session = session;
      this.remote = session.getRemote();
      this.start();
   }
   
   /**
    * Add a new message at the end of the FIFO queue and notify the push tread.
    * 
    * @param   message
    *          A binary or a text message.
    */
   public void pushMessage(Object message)
   {
      messages.offerLast(message);
      // notify
      synchronized (messages)
      {
         messages.notify();
      }
   }
   
   /**
    * 
    */
   @Override
   public void run()
   {
      while(running)
      {
         try
         {
            // are messages in the queue ?
            if (messages.isEmpty())
            {
               // no messages on the queue. wait.
               synchronized (messages)
               {
                  messages.wait();
               }                  
            }
            
            // get a message from queue
            Object message = messages.pollFirst();

            // send message
            if (session.isOpen())
            {
               // test message type
               if (message instanceof ByteBuffer)
               {
                  remote.sendBytes((ByteBuffer) message);
               }
               else if (message instanceof String)
               {
                  remote.sendString((String) message);
               }
            }
         }
         catch (Exception e)
         {
            LOG.logp(Level.INFO,
                     "PushMessagesWorker.run",
                     "",
                     LogHelper.generate("Interrupted!"),
                     e);      
         }
      }
   }

   /**
    * Kill thread and wait to die.
    */
   public void stopPushing()
   {
      running = false;
      this.interrupt();

      // wait for thread to exit      
      try
      {
         this.join(PushMessagesWorker.WAIT_TO_DIE);
      }
      catch (InterruptedException e)
      {
         LOG.logp(Level.INFO,
                  "PushMessagesWorker.stopPushing",
                  "",
                  LogHelper.generate("Interrupted!"),
                  e);      
      }
   }   
}
