/*
** Module   : tty.screen.js
** Abstract : screen management.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20140110 First version.
** 002 MAG 20140204 Implements text attributes.
** 003 MAG 20140211 Implements COLOR.
** 004 MAG 20140217 Minor changes.
** 005 MAG 20140219 Performances improvements.
**                  Double buffered implementation.
*/

tty.screen = (function()
{
   /* exports object */
   var my = {};
   
   /* used to calculate character width */
   var calibration = 'THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG';

   /* canvas */
   var canvas, mainCanvas;
   
   /* canvas 2d context */
   var ctx, mainCtx;
   
   /* canvas font */
   var font, bold;
   
   /* screen dimensions */
   var rows, cols;      
   
   /* font width and height */
   var dx, dy;

   /* color palette */
   my.palette;
   
   /*
   ** Module   : video
   ** Abstract : screen video memory
   */
   var video = (function()
   {
      /* exports object */
      var my = {};      
      
      /* video memory */
      var memo;      
      
      /* initialize video memory */
      my.init = function(size, color)
      {
         my.color = color;         
         memo = new Array(size.rows);
         for (i = 0;i < size.rows;i++) 
         {
            memo[i] = new Array(size.cols);
            for (j = 0;j < size.cols;j++)
            {
               var cell = {'c':' ','f':'white','b':'black','a':0,'k':0};
               cell.f = color.f;
               cell.b = color.b;
               memo[i][j] = cell;
            }
         }
      };

      /* clear video memory */
      my.clear = function(rows, cols, color)
      {
         for (i = 0;i < rows;i++) 
         {
            for (j = 0;j < cols;j++)
            {
               var v = memo[i][j]; 
               v.c = ' ';
               v.f = color.f;
               v.b = color.b;
               v.a = 0;
               v.k = 0;
            }
         }
      };

      /* get cell value */
      my.get = function(row, col)
      {
         return memo[row][col];
      };

      /* set cell value*/
      my.put = function(row, col, cell)
      {
         memo[row][col] = cell;
      };

      /* set cell value*/
      my.putChar = function(c, row, col, a, k)
      {
         var v = memo[row][col]; 
         v.c = c;
         v.a = a;
         v.k = k;
      };
      
      /* exports */
      return my;
   })();

   /*
   ** Module   : cursor 
   ** Abstract : screen cursor management
   */
   var cursor = (function(screen)
   {
      /* exports object */
      var my = {};
      
      /* visibility */
      my.visible = true;

      /* hide cursor */
      my.hide = function()
      {
         screen.drawCursor(false);
      };
      
      /* show cursor */
      my.show = function()
      {
         screen.drawCursor(my.visible);
      };

      /* advance cursor to next position */
      my.next = function(rows, cols)
      {
         my.col++;
         if (my.col === cols)
         {
            my.col = 0;
            my.row++;
            if (my.row === rows)
            {
               my.row = 0;
            }
         }
      };

      /* advance cursor to previous position */
      my.back = function(rows, cols)
      {
         my.col--;
         if (my.col < 0)
         {
            my.col = my.col + cols;
            my.row--;
            if (my.row < 0)
            {
               my.row = my.row + rows;
            }
         }
      };
      
      /* set cursor position */
      my.setPosition = function(row, col)
      {
         my.hide();
         my.row = row;
         my.col = col;
         my.show();
      };
      
      /*
      ** Module   : blinking 
      ** Abstract : screen cursor blinking management
      */
      my.blinking = (function(cursor, screen) 
      {
         /* exports object */
         var my = {};
         /* timer */
         var timer;

         /* stop blinking */
         my.off = function()
         {
            window.clearInterval(timer);
         };

         /* start blinking */
         my.on = function()
         {
            var blink = true;
            timer = window.setInterval(function()
            {
               screen.drawCursor(blink & cursor.visible);
               blink = !blink;
            }, 500);
         };

         /* exports */
         return my;
      }(my, screen));
      
      /* initialize cursor */
      my.init = function(cursor)
      {
         my.row = 0;
         my.col = 0;
         my.type = cursor.type;
         my.visible = cursor.visible;
         if (cursor.blinking) 
         {
            my.blinking.on();
         }
      };
      
      /* exports */
      return my;
   })(my);
   
   /* draw cursor */
   my.drawCursor = function(on)
   {
      var v = video.get(cursor.row, cursor.col);
                       
      var x = cursor.col * dx;
      var y = cursor.row * dy;
      
      if (on)
      {
         if (cursor.type === 'solid')
         {         
            v.f = video.color.b;
            v.b = video.color.f;
            my.printCell(cursor.row, cursor.col, v);
         } 
         else 
         {
            my.printCharAt(cursor.row, cursor.col);         
            // draw line cursor           
            ctx.fillStyle = video.color.f;
            ctx.fillRect(x, y + dy - 4, dx, 4);
         }
      }
      else
      {
         my.printCharAt(cursor.row, cursor.col);
      }
      draw();
   };
   
   /* set cell colors */
   var setColor = function(row, col)
   {
      var v = video.get(row, col);

      if(v.k === 0)
      {
         v.f = video.color.f;
         v.b = video.color.b;
      }
      else
      {
         // check for colors
         var fg = ((v.k >>> 4) & 0x07) - 1;
         var bg = (v.k & 0x07) - 1;
      
         v.f = my.palette[fg];
         v.b = my.palette[bg];
      }

      return v;
   };

   /* print cell */
   my.printCell = function(row, col, v)
   {
      ctx.font = (v.a & 0x02) ? bold : font;                 
      ctx.textBaseline = 'top';
            
      var x = col * dx;
      var y = row * dy;
      
      /* reverse */
      if (v.a & 0x08)
      {
         ctx.fillStyle = v.f;
         ctx.fillRect(x, y, dx, dy);
         ctx.fillStyle = v.b;
      }
      else
      {
         ctx.fillStyle = v.b;
         ctx.fillRect(x, y, dx, dy);
         ctx.fillStyle = v.f;
      }
      
      /* browser rendering correction */
      ctx.fillText(v.c, x, y - my.cy + 1);     
      
      /* underline */
      if (v.a & 0x04) 
      {
         ctx.fillStyle = v.f;
         ctx.fillRect(x, y + dy - 1, dx, 1);
      }   
   };

   /* print cell at (row, col) */
   my.printCharAt = function(row, col)
   {
      var v = setColor(row, col);
      my.printCell(row, col, v);            
   };

   /* print char at (row, col) */
   my.printChar = function(c, row, col, a, k)
   {
      video.putChar(c, row, col, a, k);
      my.printCharAt(row, col);
   };
   
   /* print an error message */
   my.error = function(text)
   {
      var c = cursor.col;
      var r = cursor.row;
      cursor.col=0;
      cursor.row = rows-1;
      for (var i = 0;i < text.length;i++)
      {
         my.printChar(text.charAt(i), cursor.row, cursor.col, 0, 0);
         cursor.next(rows, cols);
      }
      cursor.setPosition(r, c);
      draw();
      tty.sound.beep();
   };
   
   /* initialize canvas */
   my.init = function(cfg)
   {
      if (cfg.font.name === undefined || cfg.font.name.toLowerCase() === 'monospaced')
      {
         cfg.font.name = 'monospace';
      }
      rows = cfg.canvas.size.rows;
      cols = cfg.canvas.size.cols;    
      video.init(cfg.canvas.size, cfg.font.color);
      mainCanvas = document.getElementById(cfg.canvas.id);
      mainCtx = mainCanvas.getContext('2d');
      canvas = document.createElement('canvas');
      ctx = canvas.getContext('2d');
      font = cfg.font.size + 'px ' + cfg.font.name;
      bold = 'bold ' + font;
      ctx.font = font;
      dx = ctx.measureText(calibration).width / calibration.length;
      dy = cfg.font.size + 2;
      canvas.width = cols * dx;
      canvas.height = rows * dy;
      mainCanvas.width = canvas.width;
      mainCanvas.height = canvas.height;
      ctx.fillStyle = cfg.font.color.b;
      ctx.fillRect(0, 0, canvas.width, canvas.height);
      cursor.init(cfg.cursor);
      document.addEventListener('keydown', function(evt)
      {
         tty.keyboard.onkeydown(evt);
      }, true);
      document.addEventListener('keypress', function(evt)
      {
         tty.keyboard.onkeypress(evt);
      }, true);
      document.addEventListener('keyup', function(evt)
      {
         tty.keyboard.onkeyup(evt);
      }, true);
      if (cfg.canvas.mouse)
      {
         // mouse click
         canvas.addEventListener('click', function(evt)
         {
            var rect = canvas.getBoundingClientRect();           
            var x = evt.clientX - rect.left;
            var y = evt.clientY - rect.top;
            var c = (x / dx) | 0;
            var r = (y / dy) | 0;
            if ((r >= 0 && r < rows) && (c >= 0 && c < cols))
            {
               cursor.setPosition(r, c);
            }            
         }, true);
      }
      my.cy = tty.isFirefox ? 0 : tty.isChromium ? 3 : 2;
      draw();
      mainCanvas.focus();      
   };
   
   /* cursor next */
   my.next = function()
   {
      cursor.next(rows, cols);
   };

   /* cursor back */
   my.back = function()
   {
      cursor.back(rows, cols);
   };

   /* set cursor position */   
   my.setCursorPosition = function(x, y)
   {
      cursor.setPosition(x, y);
   };
   
   /* set cursor visibility */   
   my.setCursorStatus = function(visible)
   {
      cursor.visible = visible;
      cursor.show();
   };

   /* clear screen */   
   my.clear = function()
   {
      video.clear(rows, cols, font);
      ctx.fillStyle = font.color.b;
      ctx.fillRect(0, 0, canvas.width, canvas.height);
      cursor.setPosition(0, 0);        
      draw();
   };

   /* draw line */   
   my.drawLine = function(line)
   {
      var row = line[1];
      for(var i = 2;i < line.length; i+= 5)
      {
         var col = line[i];
         var c = String.fromCharCode(line[i+1] << 8 | line[i+2]);
         var a = line[i+3];
         var k = line[i+4];
         my.printChar(c, row, col, a, k);
      }
      draw();
   };

   /* double buffered draw */   
   var draw = function()
   {
      mainCtx.drawImage(canvas, 0, 0);      
   };
   
   /* exports */
   return my;       
}());
